"""Fetch data from agile sources and return standard AgileTickets."""

from dateutil.parser import parse
from jira import JIRA

from .models import AgileTicket


class BaseFetcher(object):
    """Base class for Fetchers."""

    def fetch():
        """Method invoked to fetch.

        Args:
            None
        Returns:
            None
        Raises:
            NotImplementedError
        """
        raise NotImplementedError  # pragma: no cover


def convert_jira_issue(issue):
    """
    Convert a JIRA issue into a AgileTicket.

    Args:
        issue (Issue): A jira.Issue instance

    Returns:
        An AgileTicket instance

    Raises:
        None
    """
    try:
        ttype = issue.fields.issuetype.name
    except AttributeError:
        ttype = "Ticket"

    t = AgileTicket(issue.key, ttype=ttype)

    t.title = issue.fields.summary
    t.created_at = parse(issue.fields.created)
    t.updated_at = parse(issue.fields.updated)
    t.flow_log.append(
        dict(
            entered_at=t.created_at,
            state=str("Created"),
        )
    )

    for history in issue.changelog.histories:
        for item in history.items:
            if item.field == 'status':
                t.flow_log.append(
                    dict(
                        entered_at=parse(history.created),
                        state=str(item.toString)
                    )
                )

    return t


class JIRAFetcher(BaseFetcher):
    """Fetch data from JIRA and transform it into AgileTickets.

    Attributes:
        jira_kwargs (dict): The arguments passed to the JIRA instance.
        auth_kwargs (dict): The authentication kwargs passed to the JIRA instance
    """

    BASIC_AUTH_KEYS = ['username', 'password']
    OAUTH_KEYS = ['access_token', 'access_token_secret', 'consumer_key', 'key_cert']

    def __init__(self, url, auth, filter_id, max_results=999, jira_kwargs=None):
        """Create JIRAFetcher.

        Args:
            url (str): Fully qualified URL include http/https scheme to your JIRA instance
            auth (dict): Dictionary of authentication credentials. Either username/password keys
                for basic auth OR access_token/access_token_secret/consumer_key/key_cert for OAuth
            filter_id (int): The JIRA filter ID of results you wish to fetch
            max_results (Optional[int]): Number of results to fetch, defaults to 999
            jira_kwargs (Optional[dict]): Additional kwargs passed to the jira.JIRA class
                at instance creation

        Returns:
            JIRAFetcher: instance

        Raises:
            TypeError: When neither Basic nor Oauth keys were provided in the auth dict
        """
        self._url = url
        self._auth = auth
        self.auth_kwargs = {}
        self._validate_auth()
        self._filter_id = int(filter_id)
        self.jira_kwargs = jira_kwargs or {}
        self.jira_kwargs.update(self.auth_kwargs)
        self._max_results = max_results

    def _validate_auth(self):
        if set(self._auth.keys()) <= set(self.BASIC_AUTH_KEYS):
            self.auth_kwargs = dict(basic_auth=(self._auth['username'], self._auth['password']))
        elif set(self._auth.keys()) <= set(self.OAUTH_KEYS):
            self.auth_kwargs = dict(oauth=self._auth)
        else:
            raise TypeError("Neither %s nor %s found in auth parameter" % (self.BASIC_AUTH_KEYS, self.OAUTH_KEYS))

    # TODO: Needs tests
    def fetch(self, jira_klass=JIRA):
        """Fetch data and return AgileTickets.

        Args:
            jira_klass (Optional[JIRA]): jira.JIRA compatible class to be used for a JIRA connection

        Returns:
            list: List of AgileTicket instances or empty list

        Raises:
            None
        """
        j = jira_klass(
            server=self._url,
            **self.jira_kwargs
        )
        search_string = "filter={filter_id}".format(filter_id=self._filter_id)
        issues = j.search_issues(search_string, maxResults=self._max_results, expand="changelog")
        tickets = []
        for i in issues:
            tickets.append(convert_jira_issue(i))
        return tickets



import csv
import re

import gspread

from oauth2client.service_account import ServiceAccountCredentials


class CSVWriter(object):
    def write(self, report, destination):
        """Write a CSV version of the report.
        Arguments:
            desitnation (file-like object): The file-like object where the data should be written to.
            report (Report): The Report instance that should be written to the destination.
        Returns:
            None
        """
        writer = csv.writer(destination)
        for row in report.table:
            writer.writerow(row)


class GSheetWriter(object):
    """Writes reports to Google Spreadsheets.
    Arguments:
        keyfile_name (str): The path to a JSON file with the service account credentials you want to use.
    """

    CREDENTIAL_CLASS = ServiceAccountCredentials
    DRIVER_MODULE = gspread
    COLUMN_OPTIONS = 'A B C D E F G H I J K L M N O P Q R S T U V W X Y Z'.split(' ')

    def __init__(self, keyfile_name="client_secret.json"):
        self.keyfile_name = keyfile_name
        self.scope = ['https://spreadsheets.google.com/feeds']

    @property
    def credentials(self):
        """Load the JSON and return ServiceAccountCredentials
        Returns:
            Credentials: The oauth2client compatible credentials object based on the JSON.
        """
        credentials = self.CREDENTIAL_CLASS.from_json_keyfile_name(self.keyfile_name, self.scope)
        return credentials

    @property
    def driver(self):
        return self.DRIVER_MODULE.authorize(self.credentials)

    def get_datasheet(self, doc, name):
        """Finds (or creates) the worksheet in the supplied google spreadsheet.
        Arguments:
            doc (Spreadsheet): The spreadsheet instance where you'd like the worksheet to exist.
            name (str default="Data"): The name of the worksheet you'd like to exist.
        Returns:
            Worksheet: The worksheet requested.
        """
        try:
            data_sheet = doc.worksheet(name)
        except gspread.exceptions.WorksheetNotFound:
            data_sheet = doc.add_worksheet(name, 1, 1)
        return data_sheet

    def clear_sheet(self, sheet, rows, cols):
        """Deletes all data on the supplied sheet by resizing it.
        Arguments:
            sheet (Worksheet): The worksheet you want cleared.
            rows (int): The number of rows the cleared out sheet should end up with.
            cols (int): The number of rows the cleared out sheet should end up with.
        Returns:
            None
        """
        sheet.resize(rows, cols)
        cell_list = sheet.findall(re.compile(".*", re.DOTALL))
        for cell in cell_list:
            cell.value = ""
        sheet.update_cells(cell_list)

    def select_range(self, sheet, data):
        """Returns the cells that will be modified to match the new data.
        Arguments:
            sheet (Worksheet): The worksheet you want to put data into.
            data (array): The array of data you want placed into the sheet.
        Returns:
            list[Cell]: The list of gspread.Cells that should be replaced to match the data.
        """
        start_cell = "A1"  # Always and forever
        end_cell = "{}{}".format(
            self.COLUMN_OPTIONS[len(data[0]) - 1],
            len(data)
        )
        return sheet.range("{}:{}".format(start_cell, end_cell))

    def update_cells(self, cells, data):
        """Updates the cells, in batch, with the data provided.
        Arguments:
            cells (list[Cell]): The list of gspread.Cells that should be replaced to match the data.
            data (array): The array of data you want placed into the sheet.
        Returns:
            None
        """
        for i in range(0, len(cells)):
            cell = cells[i]
            row, col = cell.row, cell.col
            row_index = row - 1
            col_index = col - 1
            try:
                value = data[row_index][col_index]
                cell.value = value
            except IndexError:  # Leave as is, not in data
                pass

        return cells

    def batch_update_sheet(self, sheet, data):
        """Clear a sheet and append the data to it.
        Arguments:
            sheet (Worksheet): The worksheet you want to modify
            data (array): The data you want the worksheet to contain
        Returns:
            None
        """
        self.clear_sheet(sheet, len(data), len(data[0]))
        cells = self.select_range(sheet, data)
        cells = self.update_cells(cells, data)
        sheet.update_cells(cells)

    def write(self, report, doc_name, sheet_name):
        doc = self.driver.open(doc_name)
        sheet = self.get_datasheet(doc, sheet_name)
        self.batch_update_sheet(sheet, report.table)



"""Make reports from data."""

from collections import namedtuple
from datetime import datetime

from dateutil.relativedelta import relativedelta
from dateutil.tz import tzutc

from numpy import histogram, array, arange, percentile, round

Report = namedtuple("Report", ["table", "summary"])


class Reporter(object):
    """Base class for Reporters.

    Attributes:
        title (unicode): The name of the report
        start_date (datetime): The starting range of the report.
        end_date (datetime): The ending range of the report.
    """

    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6

    def __init__(self, title, start_date=None, end_date=None):
        self.title = title
        self.start_date = start_date
        self.end_date = end_date
        super().__init__()

    @property
    def start_date(self):
        return self.valid_start_date(self._start_date)

    @start_date.setter
    def start_date(self, value):
        if value and value.tzinfo is None:
            value = value.replace(tzinfo=tzutc())
        self._start_date = value

    @property
    def end_date(self):
        return self.valid_end_date(self._end_date)

    @end_date.setter
    def end_date(self, value):
        if value and value.tzinfo is None:
            value = value.replace(tzinfo=tzutc())
        self._end_date = value

    def valid_start_date(self, target_date):
        """Returns a date that is valid for the start of the report.
        Arguments:
            target_date (datetime): The date you'd like examined
        Returns:
            datetime: A datetime made valid for the report based on the target_date argument.
        """
        return target_date

    def valid_end_date(self, target_date):
        """Returns a date that is valid for the end of the report.
        Arguments:
            target_date (datetime): The date you'd like examined
        Returns:
            datetime: A datetime made valid for the report based on the target_date argument.
        """
        return target_date

    def walk_back_to_weekday(self, target_date, day):
        """Returns the nearest date that predates the target_date.
        Arguments:
            target_date (datetime): The date to start with.
            day (int): An integer between 0 (Monday) and 6 (Sunday)
        Returns:
            datetime: The nearest date that predates the target_date for the given day.
        """
        while target_date.weekday() != day:
            target_date = target_date - relativedelta(days=1)
        return target_date

    def walk_forward_to_weekday(self, target_date, day):
        """Returns the nearest date that postdates the target_date.
        Arguments:
            target_date (datetime): The date to start with.
            day (int): An integer between 0 (Monday) and 6 (Sunday)
        Returns:
            datetime: The nearest date that postdates the target_date for the given day.
        """
        while target_date.weekday() != day:
            target_date = target_date + relativedelta(days=1)
        return target_date

    def filter_on_ended(self, issues):
        """Returns issues that were ended between the instances start/end dates.
        Arguments:
            issues (list[AnalyzedAgileTicket]): List of issues to be filtered_issues
        Return:
            list[AnalyzedAgileTicket]: List of issues that match.
        """
        filtered_issues = [i for i in issues if i.ended and (i.ended['entered_at'] >= self.start_date and i.ended['entered_at'] <= self.end_date)]
        return filtered_issues

    def filter_on_committed(self, issues):
        """Returns issues that were committed to between the instances start/end dates.
        Arguments:
            issues (list[AnalyzedAgileTicket]): List of issues to be filtered_issues
        Return:
            list[AnalyzedAgileTicket]: List of issues that match.
        """
        filtered_issues = [i for i in issues if i.ended and (i.committed['entered_at'] >= self.start_date and i.committed['entered_at'] <= self.end_date)]
        return filtered_issues

    def starts_of_weeks(self):
        """Return a list of dates that start each week between start_date and end_date."""
        week_starting = self.walk_back_to_weekday(self.start_date.date(), self.SUNDAY)
        while week_starting <= self.end_date.date():
            yield week_starting
            week_starting += relativedelta(days=7)

    def filter_issues(self, issues):
        raise NotImplementedError

    def report_on(self, issues, header=True):
        raise NotImplementedError


class CreatedReporter(Reporter):
    """Generates a report listing counts for all the types of tickets created each week."""
    def valid_start_date(self, target_date):
        """Ensure we start on a Sunday."""
        target_date = super().valid_start_date(target_date)
        return self.walk_back_to_weekday(target_date, self.SUNDAY)

    def valid_end_date(self, target_date):
        """Ensure we end on a Sunday."""
        target_date = super().valid_end_date(target_date)
        return self.walk_forward_to_weekday(target_date, self.SATURDAY)

    def filter_issues(self, issues):
        """Ignore issues completed outside the start/end range."""
        return self.filter_on_committed(issues)

    def _issues_for_week(self, week_start, issues):
        week_end = self.walk_forward_to_weekday(week_start, self.SATURDAY)
        week_start = datetime(week_start.year, week_start.month, week_start.day, 0, 0, 0, tzinfo=tzutc())
        week_end = datetime(week_end.year, week_end.month, week_end.day, 11, 59, 59, tzinfo=tzutc())
        return [i for i in issues if i.committed['entered_at'] >= week_start and i.committed['entered_at'] <= week_end]

    def report_on(self, issues, header=True):
        """Generate a report, one row per week, with counts for each ticket type."""
        issues = self.filter_issues(issues)
        r = Report(
            table=[],
            summary=dict(
                title=self.title,
                start_date=self.start_date,
                end_date=self.end_date
            )
        )

        ticket_types = list(set([issue.type for issue in issues]))
        ticket_types.sort()
        headers = ["Week", ]
        headers.extend(ticket_types)
        r.table.append(headers)

        for sunday in self.starts_of_weeks():
            this_weeks_issues = self._issues_for_week(sunday, issues)
            row = [sunday, ]
            for ttype in ticket_types:
                row.append(len([i for i in this_weeks_issues if i.type == ttype]))
            r.table.append(row)

        return r


class TicketReporter(Reporter):
    """Generates a report listing all the tickets that match the filter critiera."""
    def valid_start_date(self, target_date):
        """Ensure we start on a Sunday."""
        target_date = super().valid_start_date(target_date)
        return self.walk_back_to_weekday(target_date, self.SUNDAY)

    def valid_end_date(self, target_date):
        """Ensure we end on a Sunday."""
        target_date = super().valid_end_date(target_date)
        return self.walk_forward_to_weekday(target_date, self.SATURDAY)

    def filter_issues(self, issues):
        """Ignore issues completed outside the start/end range."""
        return self.filter_on_ended(issues)

    def report_on(self, issues, header=True):
        """Generate a report, one row per issue, with details."""
        issues = self.filter_issues(issues)
        issues.sort(key=lambda x: x.ended['entered_at'])
        r = Report(
            table=[],
            summary=dict(
                title=self.title,
                start_date=self.start_date,
                end_date=self.end_date
            )
        )

        r.table.append(
            ["Key", "Title", "Lead Time", "Cycle Time", "Commit State", "Commit At", "Start State", "Start At", "End State", "End At"],
        )
        for i in issues:
            row = [
                i.key,
                i.title,
                i.lead_time,
                i.cycle_time,
                i.committed['state'],
                i.committed['entered_at'],
                i.started['state'],
                i.started['entered_at'],
                i.ended['state'],
                i.ended['entered_at'],
            ]
            r.table.append(row)

        return r


class SLAReporter(Reporter):
    """Generates a report showing the number of tickets per week that exceeded their SLA."""
    def valid_start_date(self, target_date):
        """Ensure we start on a Sunday."""
        target_date = super().valid_start_date(target_date)
        return self.walk_back_to_weekday(target_date, self.SUNDAY)

    def valid_end_date(self, target_date):
        """Ensure we end on a Sunday."""
        target_date = super().valid_end_date(target_date)
        return self.walk_forward_to_weekday(target_date, self.SATURDAY)

    def filter_issues(self, issues):
        """Ignore issues completed outside the start/end range."""
        return self.filter_on_ended(issues)

    def _issues_for_week(self, week_start, issues):
        week_end = self.walk_forward_to_weekday(week_start, self.SATURDAY)
        week_start = datetime(week_start.year, week_start.month, week_start.day, 0, 0, 0, tzinfo=tzutc())
        week_end = datetime(week_end.year, week_end.month, week_end.day, 11, 59, 59, tzinfo=tzutc())
        return [i for i in issues if i.ended['entered_at'] >= week_start and i.ended['entered_at'] <= week_end]

    def report_on(self, issues, sla_config={}, header=True):
        r = Report(
            table=[],
            summary=dict(
                title=self.title,
                start_date=self.start_date,
                end_date=self.end_date
            )
        )

        issues = self.filter_issues(issues)
        ticket_types = list(set([issue.type for issue in issues]))
        ticket_types.sort()
        headers = ["Week", ]
        headers.extend(ticket_types)
        r.table.append(headers)

        for sunday in self.starts_of_weeks():
            this_weeks_issues = self._issues_for_week(sunday, issues)
            row = [sunday, ]
            for ttype in ticket_types:
                row.append(len([i for i in this_weeks_issues if i.type == ttype and i.lead_time > sla_config[ttype]]))
            r.table.append(row)

        return r


class LeadTimeDistributionReporter(Reporter):
    """Generates lead time distribution report for tickets completed in a single date range."""

    def valid_start_date(self, target_date):
        """Ensure we start on a Sunday."""
        target_date = super().valid_start_date(target_date)
        return self.walk_back_to_weekday(target_date, self.SUNDAY)

    def valid_end_date(self, target_date):
        """Ensure we end on a Sunday."""
        target_date = super().valid_end_date(target_date)
        return self.walk_forward_to_weekday(target_date, self.SATURDAY)

    def filter_issues(self, issues):
        """Ignore issues completed outside the start/end range."""
        return self.filter_on_ended(issues)

    def report_on(self, issues, header=True):
        """Generate a report object with a lead time histogram."""
        r = Report(
            table=[],
            summary=dict(
                title=self.title,
                start_date=self.start_date,
                end_date=self.end_date
            )
        )
        r.table.append(["Lead Time", "Tickets"])
        filtered_issues = self.filter_issues(issues)

        if filtered_issues:
            lead_times = [i.lead_time for i in filtered_issues]
            lead_times = array(lead_times)
            hist_values, hist_bins = histogram(
                lead_times,
                bins=arange(0, max(lead_times) + 2, 1)
            )

            for i in range(0, len(hist_values)):
                if i == 0:
                    continue
                row = [hist_bins[i], hist_values[i]]
                r.table.append(row)
        return r


class TimePercentileReporter(Reporter):
    """Report on Time Percentiles, calculated over a configurable (default 4) week period.
    Attributes:
        title (unicode): The name of the report
        start_date (datetime): The starting range of issues for the report.
        end_date (datetime): The ending range of issues for the report.
        num_weeks (int): The number of weeks you'd like reported on. Default: 4
    """

    def __init__(self, title, time_attr, start_date=None, end_date=None, num_weeks=4):
        super().__init__(title, start_date, end_date)
        self.num_weeks = num_weeks
        self.time_attr = time_attr

    def valid_start_date(self, target_date):
        target_date = super().valid_start_date(target_date)
        target_date = self.walk_back_to_weekday(target_date, self.SUNDAY)
        return target_date

    def valid_end_date(self, target_date):
        target_date = super().valid_end_date(target_date)
        target_date = self.walk_forward_to_weekday(target_date, self.SATURDAY)
        return target_date

    def filter_issues(self, issues):
        return self.filter_on_ended(issues)

    def _times_for_week(self, week_start, issues):
        week_end = self.walk_forward_to_weekday(week_start, self.SATURDAY)
        week_start = datetime(week_start.year, week_start.month, week_start.day, 0, 0, 0, tzinfo=tzutc())
        week_end = datetime(week_end.year, week_end.month, week_end.day, 11, 59, 59, tzinfo=tzutc())
        return [getattr(i, self.time_attr) for i in issues if i.ended['entered_at'] >= week_start and i.ended['entered_at'] <= week_end]

    def report_on(self, issues, header=True):
        r = Report(
            table=[],
            summary=dict(
                title=self.title,
                start_date=self.start_date,
                end_date=self.end_date,
                num_weeks=self.num_weeks)
        )
        if header:
            headers = ["Week", "50th", "75th", "95th"]
            r.table.append(headers)

        issues = self.filter_issues(issues)

        percentiles_by_week = []
        moving_sample = []
        for sunday in self.starts_of_weeks():
            moving_sample.append(self._times_for_week(sunday, issues))
            # Lets limit our percentile calc to the past 4 weeks
            if len(moving_sample) > 4:
                moving_sample.pop(0)

            samples = []
            for sample_set in moving_sample:
                samples.extend(sample_set)

            try:
                the_50th = int(round(percentile(samples, 50)))
                the_75th = int(round(percentile(samples, 75)))
                the_95th = int(round(percentile(samples, 95)))
            except IndexError:
                the_50th, the_75th, the_95th = 0, 0, 0

            percentiles_by_week.append([sunday, the_50th, the_75th, the_95th])

        for row in percentiles_by_week[-self.num_weeks:]:
            r.table.append(row)

        return r


class LeadTimePercentileReporter(TimePercentileReporter):
    def __init__(self, title, start_date=None, end_date=None, num_weeks=4):
        super(LeadTimePercentileReporter, self).__init__(title, "lead_time", start_date=start_date, end_date=end_date, num_weeks=num_weeks)


class CycleTimePercentileReporter(TimePercentileReporter):
    def __init__(self, title, start_date=None, end_date=None, num_weeks=4):
        super(CycleTimePercentileReporter, self).__init__(title, "cycle_time", start_date=start_date, end_date=end_date, num_weeks=num_weeks)


class ThroughputReporter(Reporter):
    """Generate throughput reports.

    Attributes:
        title (unicode): The name of the report
        period (unicode): The interval you'd like, one of daily, weekly, monthly,
        start_date (datetime): The starting range of the report.
        end_date (datetime): The ending range of the report.
    """

    def __init__(self, title, period=None, start_date=None, end_date=None):
        self.title = title
        self.period = period
        self.start_date = start_date
        self.end_date = end_date
        super().__init__(title, start_date, end_date)

    def valid_start_date(self, target_date):
        target_date = super().valid_start_date(target_date)
        if self.period == "weekly":
            target_date = self.walk_back_to_weekday(target_date, self.SUNDAY)
        return target_date

    def valid_end_date(self, target_date):
        target_date = super().valid_end_date(target_date)
        if self.period == "weekly":
            target_date = self.walk_forward_to_weekday(target_date, self.SATURDAY)
        return target_date

    def _count_by_week(self, issues):
        counted_by_week = {}
        for week_starting in self.starts_of_weeks():
            week_end = week_starting + relativedelta(days=6)
            counted_by_week[week_starting] = len(
                [i for i in issues if i.ended['entered_at'].date() >= week_starting and i.ended['entered_at'].date() <= week_end]
            )

        return counted_by_week

    def filter_issues(self, issues):
        return self.filter_on_ended(issues)

    def report_on(self, issues, header=True):
        r = Report(
            table=[],
            summary=dict(
                title=self.title,
                period=self.period,
                start_date=self.start_date,
                end_date=self.end_date
            )
        )
        r.table.append(["Week", "Completed"])
        filtered_issues = self.filter_issues(issues)
        counted_by_week = self._count_by_week(filtered_issues)

        weeks = list(counted_by_week.keys())
        weeks.sort()
        for week in weeks:
            r.table.append([week, counted_by_week[week]])

        return r



"""Analyzers decorate AgileTickets with contextual information.

Analyzers look at tickets through the lens of "this is my start state", or
"this is what defects look like" and modify AgileTickets to contain information
based on that context like "ended_at", "commited_at", "started_at", etc.
"""

from .models import AnalyzedAgileTicket


class MissingPhaseInformation(Exception):
    """Raise when a ticket is missing information for a phase.

    Arguments:
        message (unicode): Human readable string describing the exception.
        phase (unicode): The phase that no state could be found for.
        state_list (list[unicode]): List of states that were included in the phase.

    Attritbutes:
        message (unicode): Human readable string describing the exception.
        phase (unicode): The phase that no state could be found for.
        state_list (list[unicode]): List of states that were included in the phase.
    """

    def __init__(self, message, phase, state_list):
        """Create the exception."""
        self.message = message
        self.phase = phase
        self.state_list = state_list
        super(Exception, self).__init__(message)


class PartialDateAnalyzer(object):
    """Analyze Tickets that might not have been started or completed.

    Attributes:
        commit_state (list[str]): The list of names of the state when work was committed to.
        start_state (list[str]): The list of names of the state when work was started.
        end_state (list[str]): The list of names of the state when work was completed.
    """
    NEWEST_DATE = 'newest'
    OLDEST_DATE = 'oldest'

    def __init__(self, commit_states, start_states, end_states):
        """Create instances."""
        self.end_states = end_states
        self.commit_states = commit_states
        self.start_states = start_states
        super().__init__()

    @property
    def states_context(self):
        """Enumerate the states that match the phases of an analyzed ticket."""
        return {
            u'committed': self.commit_states,
            u'started': self.start_states,
            u'ended': self.end_states,
        }

    def _find_entered_at(self, state_list, ticket, strategy):
        entry = dict(state=None, entered_at=None)
        entries = []
        for state_name in state_list:
            for log in ticket.flow_log:
                if log['state'] == state_name:
                    entries.append(log)
            if len(entries) > 0:
                break

        if len(entries) > 0:
            if strategy == self.NEWEST_DATE:
                entry = entries[-1]
            else:
                entry = entries[0]
        return entry['state'], entry['entered_at']

    def analyze(self, tickets, strategy=None):
        """Return a list of AnalyzedAgileTicket.

        Arguments:
            tickets (list[AgileTicket]): The list of tickets to be analyzed
            strategy (analyzer.OLDEST_DATE | analyzer.NEWEST_DATE): Which date to pick when a ticket entered a state multiple times

        Returns:
            list[AnalyzedAgileTicket]: The list of tickets
        """
        if strategy is None:
            strategy = self.OLDEST_DATE

        analyzed_tickets = []
        ignored_tickets = []
        for ticket in tickets:
            analyzed_tickets.append(self.analyze_ticket(ticket, strategy))
        return analyzed_tickets, ignored_tickets

    def analyze_ticket(self, ticket, strategy):
        """Convert a single AgileTicket into an AnalyzedAgileTicket.

        Arguments:
            ticket (AgileTicket): The AgileTicket under consideration
            strategy (analyzer.OLDEST_DATE | analyzer.NEWEST_DATE): Which date to pick when a ticket entered a state multiple times

        Returns:
            AnalyzedAgileTicket
        """

        kwargs = {
            "key": ticket.key,
            "ttype": ticket.type,
            "title": ticket.title,
        }

        for phase, state_list in self.states_context.items():
            state, datetime = self._find_entered_at(state_list, ticket, strategy)
            kwargs[phase] = dict(state=state, entered_at=datetime)
        return AnalyzedAgileTicket(**kwargs)


class DateAnalyzer(PartialDateAnalyzer):
    """Analyze Tickets for cycle data.

    Attributes:
        commit_state (list[str]): The list of names of the state when work was committed to.
        start_state (list[str]): The list of names of the state when work was started.
        end_state (list[str]): The list of names of the state when work was completed.
    """

    def analyze(self, tickets, strategy=None):
        """Return a list of AnalyzedAgileTicket.

        Arguments:
            tickets (list[AgileTicket]): The list of tickets to be analyzed
            strategy (analyzer.OLDEST_DATE | analyzer.NEWEST_DATE): Which date to pick when a ticket entered a state multiple times

        Returns:
            list[AnalyzedAgileTicket]: The list of tickets
        """
        if strategy is None:
            strategy = self.OLDEST_DATE

        analyzed_tickets = []
        ignored_tickets = []
        for ticket in tickets:
            try:
                analyzed_tickets.append(self.analyze_ticket(ticket, strategy))
            except MissingPhaseInformation as e:
                ignored_tickets.append(dict(ticket=ticket, phase=e.phase, state_list=e.state_list))
        return analyzed_tickets, ignored_tickets

    def analyze_ticket(self, ticket, strategy):
        """Convert a single AgileTicket into an AnalyzedAgileTicket.

        Arguments:
            ticket (AgileTicket): The AgileTicket under consideration
            strategy (analyzer.OLDEST_DATE | analyzer.NEWEST_DATE): Which date to pick when a ticket entered a state multiple times

        Returns:
            AnalyzedAgileTicket
        """

        kwargs = {
            "key": ticket.key,
            "ttype": ticket.type,
            "title": ticket.title,
        }

        for phase, state_list in self.states_context.items():
            state, datetime = self._find_entered_at(state_list, ticket, strategy)
            if None in (state, datetime):
                msg = "{key} is missing flow_log information for {state_list}".format(key=ticket.key, state_list=state_list)
                raise MissingPhaseInformation(
                    msg,
                    phase,
                    state_list,
                )
            kwargs[phase] = dict(state=state, entered_at=datetime)
        return AnalyzedAgileTicket(**kwargs)



"""Data models."""

from datetime import datetime


class AgileTicket(object):
    """Abstract representation of tickets in Agile systems.

    Attributes:
        key (unicode): Unique identifier for the ticket in its system of record
        created_at (datetime): When was the ticket created
        updated_at (datetime): When was the ticket last updated
        type (str): The kind of ticket this is: Bug, Epic, Story, etc.

    Optional Attributes:
        title (unicode): The title of the ticket
        type (unicode): A label of the type of ticket: Story, Epic, Defect
    """

    def __init__(self, key, title="", ttype="Ticket"):
        """Init an AgileTicket.

        Args:
            key (str): A unique identifier for this ticket in the system of record
        """
        self.key = str(key)
        self.title = title
        self.created_at = None
        self.updated_at = None
        self.type = ttype
        self._flow_log = FlowLog()

    @property
    def flow_log(self):
        """FlowLog[dict].

        A list of dicts guaranteed to have the following:
            entered_at (datetime): When the ticket entered the state
            state (unicode): The name of the state the ticket entered
        """
        return self._flow_log


class FlowLog(list):
    """List subclass enforcing dictionaries with specific keys are added to it."""

    def append(self, value):
        """Add items to the list.

        Args:
            value (dict): Must contain an entered_at and state key.

        Returns:
            None

        Raises:
            TypeError: Flow log items must have a 'entered_at' datetime and a 'state' string.
        """
        try:
            ('entered_at', 'state') in value.keys()
        except AttributeError:
            raise TypeError("Flow log items must have a 'entered_at' datetime and a 'state' string. Got: {value}".format(value=value))

        entered_at = value['entered_at']
        try:
            datetime.now(entered_at.tzinfo) - entered_at
        except (AttributeError, TypeError) as e:
            msgvars = dict(
                val_type=type(entered_at),
                val=entered_at,
                exc=str(e)
            )
            raise TypeError("Flow log items must have a entered_at datetime. Got: {val_type} / {val}, \n Exception: {exc}".format(**msgvars))

        value[u'state'] = str(value['state'])
        super(FlowLog, self).append(value)
        self.sort(key=lambda l: l['entered_at'])


class AnalyzedAgileTicket(object):
    """An AgileTicket analyzed within a certain context.

    Attributes:
        key (unicode): Unique identifier for the ticket in its system of record
        committed (dict): The state and datetime when the story was committed
        started (dict): The state and datetime when the story was started
        ended (dict): The state and datetime when the story was ended

    Optional Attributes:
        title (unicode): The title of the ticket
        type (unicode): A label of the type of ticket: Story, Epic, Defect
    """

    def __init__(
        self, key, committed, started, ended,
        title="", ttype="Ticket",
    ):
        """Create AnalyzedAgileTickets."""
        self.key = key
        self.title = title
        self.committed = committed
        self.started = started
        self.ended = ended
        self.type = ttype

    def __repr__(self):
        """Represention of the object."""
        return "{} -- Ended: {}".format(self.key, self.ended['entered_at'])

    @property
    def lead_time(self):
        """Number of days between committed and ended."""
        diff = self.ended['entered_at'] - self.committed['entered_at']
        return diff.days

    @property
    def cycle_time(self):
        """Number of days between started and ended."""
        diff = self.ended['entered_at'] - self.started['entered_at']
        return diff.days



"""Pulls data from agile systems and analyzes it."""

from .version import __version__, __author__

from .fetchers import (
    JIRAFetcher,
    convert_jira_issue
)

from .analyzers import (
    DateAnalyzer,
    PartialDateAnalyzer,
)

from .reporters import (
    ThroughputReporter,
    LeadTimeDistributionReporter,
    TicketReporter,
    LeadTimePercentileReporter,
    CycleTimePercentileReporter,
    SLAReporter,
    CreatedReporter,
)

from .writers import (
    CSVWriter,
    GSheetWriter
)

version = ".".join(map(str, __version__))


__all__ = [
    "version",
    "__version__",
    "__author__",
    "JIRAFetcher",
    "convert_jira_issue",
    "DateAnalyzer",
    "ThroughputReporter",
    "LeadTimeDistributionReporter",
    "TicketReporter",
    "CSVWriter",
    "GSheetWriter",
    "LeadTimePercentileReporter",
    "SLAReporter",
    "PartialDateAnalyzer",
    "CreatedReporter",
    "CycleTimePercentileReporter"
]



"""setup."""

from __future__ import unicode_literals
from setuptools import setup, find_packages

from pip.req import parse_requirements

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements("requirements.txt", session=False)

# reqs is a list of requirement
# e.g. ['django==1.5.1', 'mezzanine==1.4.6']
reqs = [str(ir.req) for ir in install_reqs]

description = """Pulls data from agile systems and analyzes it."""
author = "cmheisel"
README = open('README.md', 'r').read()
try:
    version = open('version.txt', 'r').read().strip()
except FileNotFoundError:
    version = "unknown"

setup(
    name='agile-analytics',
    version=version,
    description=description,
    long_description=README,
    classifiers=[
        # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='jira agile lean kanban metrics',
    author=author,
    author_email='chris@heisel.org',
    url='https://github.com/cmheisel/agile-analytics',
    license='MIT License',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    install_requires=reqs
)
